/*
 * mixer.c: 
 * Handles all the mixer related stuff, loads plugins and initializes
 * function pointers etc. 
 *
 * Copyright (C) 2001 Sakari Lehtonen (sakari@ionstream.fi)
 * 
 */

#include "umix_common.h"
#include "plugin.h"

/* internal variables */
static umixer *mixers[MAX_MIXERS];
static int curr_mixer = 0;
static int mixer_count = 0;

/* try to init all available mixers */
int init_mixers(void)
{
	int i=0;

	while(1)
	{
		if( (mixers[i] = mixer_load_drv(globalopts.driver)) == NULL )
			break;
		if( mixer_open(globalopts.devpath) == -1 )
			break;
		if( mixer_init() == -1 )
			break;
		i++;
	}

	/* failed opening any mixers */
	if( mixer_count < 1 )
		return -1;

	return 0;
}

/* try to close all open and initialized mixers in reverse order
 * they were opened */
int close_mixers(void)
{
	int i=0;

	for(i=mixer_count-1; i>=0; i--)
	{
		mixer_set_curr(i);
		mixer_close();
	}

	return 0;
}	

/* loads the mixer driver and gets the function pointers from it */
umixer *mixer_load_drv(const char *drivername)
{
	umixer *mixer;
	umixer_func *(*get_mixer_func)(void);
	nameval *lookup;
	void *handle;

	mixer = emalloc(sizeof(umixer));
	mixer->dev = emalloc(sizeof(umixer_func));
	
	/* search if the module has been already loaded */
	lookup = lookupbyname(&drivertab, drivername);
	if( lookup == NULL )
		handle = load_plugin(drivername, PLUGIN_DRIVER);
	else
		handle = (void *)lookup->val;
	if( handle == NULL )
	{
		errormsg(_("failed loading mixer plugin %s"), drivername);
		return NULL;
	}

	get_mixer_func = dlsym(handle, "get_mixer_func");
	if( get_mixer_func == NULL )
	{
		errormsg(_("%s: not a mixer plugin"), drivername);
		return NULL;
	}
	mixer->dev = get_mixer_func();

#ifdef UMIX_DEBUG
	errormsg("mixer_load_drv: loaded %s driver", drivername);
#endif

	return mixer;
}

/* opens the mixer device, mixer_load_drv must be called first */
int mixer_open(const char *path)
{
	int retval;
	char *errstr = _("failed opening mixer");

	retval = mixers[curr_mixer]->dev->mix_open(path);
	if( retval == EMIXERR )
	{
		errormsg(_("%s"), errstr);
		return -1;
	}
	else if( retval == EMIXALLOPEN )
		return -1;

	mixer_count++;
	if( mixer_count >= MAX_MIXERS )
	{
		errormsg(_("%s: maximum limit of %d mixers reached!"), 
			errstr, MAX_MIXERS);
		mixer_count--;
		return -1;
	}
	
#ifdef UMIX_DEBUG
	errormsg("mixer_open: opened device %s", path);
#endif

	return 0;
}

/* initializes the current mixer, must be first opened with mixer_open */
int mixer_init(void)
{
	const char *errstr = _("failed initializing mixer");

	if( mixer_count < 1 )
	{
		errormsg(_("%s: no mixers to init!"), errstr);
		return -1;
	}
	if( mixers[curr_mixer]->dev->mix_init() == -1 )
	{
		errormsg("%s", errstr);
		return -1;
	}

	return 0;
}

int mixer_close(void)
{
	if( mixer_count < 1 )
	{
		errormsg(_("failed closing mixer: no mixers to close!"));
		return -1;
	}

	mixers[curr_mixer]->dev->mix_close();
	free(mixers[curr_mixer]);
	mixer_count--;

#ifdef UMIX_DEBUG
	errormsg("mixer_close: closed mixer #%d", curr_mixer);
#endif

	return 0;
}

int mixer_opt_to_chan_num(const char *str)
{
	return mixers[curr_mixer]->dev->mix_opt_to_chan_num(str);
}

int mixer_get_curr(void)
{
	return curr_mixer;
}

int mixer_set_curr(int mixer)
{
	if( mixer < 0 || mixer >= mixer_count || mixer > MAX_MIXERS  )
	{
		errormsg(_("failed setting mixer to #%d: out of bounds!"), 
			mixer);
		return -1;
	}

	curr_mixer = mixer;
	mixers[curr_mixer]->dev->mix_set_curr(mixer);

	return 0;
}

void mixer_inc_curr(int delta)
{
	curr_mixer += delta;
	curr_mixer = CLAMP(curr_mixer, 0, MAX_MIXERS);
}

int chan_get_curr(void)
{
	return mixers[curr_mixer]->dev->ch_get_curr();
}

int chan_set_curr(int ch)
{
	return mixers[curr_mixer]->dev->ch_set_curr(ch);
}

void chan_inc_curr(int delta)
{
	int ch;

	ch = chan_get_curr();
	ch += delta;
	ch = CLAMP(ch, 0, mixer_get_num_chan());
	chan_set_curr(ch);
}

int chan_get_elem(void)
{
	return mixers[curr_mixer]->dev->ch_get_elem();
}

int chan_set_elem(int elem)
{
	return mixers[curr_mixer]->dev->ch_set_elem(elem);
}

int chan_get_vol(void)
{
	return mixers[curr_mixer]->dev->ch_get_vol();
}

void chan_set_vol(int vol)
{
	vol = CLAMP(vol, VOL_MIN, VOL_MAX);
	mixers[curr_mixer]->dev->ch_set_vol(vol);
}

void chan_inc_vol(int delta)
{
	int vol;

	vol = chan_get_vol();
	vol += delta;
	vol = CLAMP(vol, VOL_MIN, VOL_MAX);
	chan_set_vol(vol);
}

int chan_set_rec(int state)
{
	return mixers[curr_mixer]->dev->ch_set_rec(state);
}	

int chan_get_lr(int which)
{
	int retval;
	int orig_elem = chan_get_elem();

	if( which == CHAN_LEFT )
		chan_set_elem(ELEM_LEFT);
	else if( which == CHAN_RIGHT )
		chan_set_elem(ELEM_RIGHT);
	else if( which == CHAN_MAX )
		chan_set_elem(ELEM_BOTH);
	retval = chan_get_vol();
	chan_set_elem(orig_elem);

	return retval;
}

void chan_set_lr(int left, int right)
{
	int orig_elem = chan_get_elem();

	chan_set_elem(ELEM_LEFT);
	chan_set_vol(left);
	chan_set_elem(ELEM_RIGHT);
	chan_set_vol(right);
	chan_set_elem(orig_elem);
}

int chan_is_stereo(void)
{
	return mixers[curr_mixer]->dev->ch_is_stereo();
}

int chan_is_recsrc(void)
{
	return mixers[curr_mixer]->dev->ch_is_recsrc();
}

int chan_is_record(void)
{
	return mixers[curr_mixer]->dev->ch_is_record();
}

int chan_is_mute(void)
{
	return mixers[curr_mixer]->dev->ch_is_mute();
}

const char *mixer_get_name(void)
{
	return mixers[curr_mixer]->dev->mix_get_name();
}

const char *mixer_get_path(void)
{
	return mixers[curr_mixer]->dev->mix_get_path();
}

const char *mixer_get_driver(void)
{
	return mixers[curr_mixer]->dev->mix_get_driver();
}

int mixer_get_num_mix(void)
{
	return mixer_count;
}

int mixer_get_num_chan(void)
{
	return mixers[curr_mixer]->dev->mix_get_numchan();
}

int mixer_get_status(void)
{
	return mixers[curr_mixer]->dev->mix_get_status();
}

const char *chan_get_name(void)
{
	return mixers[curr_mixer]->dev->ch_get_name();
}

const char *chan_get_label(void)
{
	return mixers[curr_mixer]->dev->ch_get_label();
}
