/*
 * ui_ncurses.c: 
 * Ncurses user interface for Umix. Probably could use a total rewrite. But
 * hey, this works for me.
 *
 * Copyright (C) 2001 Sakari Lehtonen <sakari@ionstream.fi>
 *
 */

#include <ncurses.h>
#include "umix_common.h"
#include "ui.h"

/* the width of the channel labels */
static int LabelWidth[MAX_MIXERS];	
/* store number of channels for every mixer globally for optimization */
static int NumChan[MAX_MIXERS];
/* hack hackity, to be fixed soon */
static int ownupdate = 0;

/* the current mixer, stored here for optimization */
static int CurrMixer = 0;
/* Okay, this could use a rewrite. But I'm lazy. These are used to store
 * balances for every channel and for muting and undoing volumes */
static int *S_Vol[MAX_MIXERS];
static int *LeftBal[MAX_MIXERS];
static int *RightBal[MAX_MIXERS];

/* position constants */
enum
{
	LABELMAX	= 10,		 /* max width of labels */
	BEGX		= 0,		 /* beginning x of the layout */
	BEGY		= 2,		 /* beginning y of the layout */
	VOLCHG		= (VOL_MAX / 50),/* amount of volume change */
	BALCHG		= 8,		 /* amount of balance change */
	NUMWIDTH	= 3,		 /* width of volume numbers */
};

/* position macros */

/* width of the bars */
#define BALWIDTH	((COLS/8) + 2)
#define VOLWIDTH	(COLS - BALWIDTH - NUMWIDTH - LabelWidth[CurrMixer] - 8)

/* where the bars begin */
#define VOLBEGX		(BEGX+LabelWidth[CurrMixer])
#define VOLBEGY		(BEGY+curr_chan)
#define BALBEGX		((COLS-3) - BALWIDTH)
#define BALBEGY		(BEGY+curr_chan)
#define STATBEGX	(VOLBEGX+NUMWIDTH)

/* the current channel marker positions */
#define VOLPOSX1	(VOLBEGX+NUMWIDTH+1)
#define VOLPOSX2	(VOLPOSX1+VOLWIDTH+1)
#define BALPOSX1	(BALBEGX-1)
#define BALPOSX2	(BALBEGX+BALWIDTH+1)

#ifndef CTRL
  #define CTRL(a)       ((a) & 0x1F) 
#endif

/* settings */
enum
{
	SET_LOAD,
	SET_SAVE,
	SET_CONFIG
};

/* which border */
enum
{
	BORD_UPPER,
	BORD_LOWER
};

/* mute action */
enum
{
	M_MUTE,
	M_EXCLUDE,
	M_SAVE,
	M_NOSAVE
};

/* the characters used in drawing */
enum
{
	CHAR_SLIDER	= '*',
	CHAR_DIV	= ':',
	CHAR_EMPTY	= '.',
	CHAR_MONO	= '|'
};

/* saved cols and lines */
static int SCols = 0;
static int SLines = 0;

/* global function prototypes */
int ui_ncurs_main(void);
void ui_ncurs_exit(void);

/* internal ui_ncurses function prototypes */
static int loop(void);
static void initscreen(void);
static void cyclemixer(void);
static void mute(int action, int ch);
static void initvolumes(int save);
static void writevol(void);
static void togglerec(void);
static void updatestat(void);
static void drawslider(void);
static void drawbalance(void);
static void drawstatus(void);
static void drawall(void);
static void drawborder(int where, const char *str);
static void drawpos(int menu, int status);
static void redrawscreen(void);
static void help(void);
static void settings(int action);
static void input(char *buf, const char *prompt, int limit);
static void manualset(void);

ui_plugin ncurses_func =
{
	ui_ncurs_main,
	ui_ncurs_exit
};

ui_plugin *get_ui_func(void)
{
	return &ncurses_func;
}

int ui_ncurs_main(void)
{
	int i, j, n;
	int numchan;
	
	/* initialize curses */
	initscreen();

	/* calculate the max length of the labels */
	for(i=mixer_get_num_mix()-1; i>=0; i--)
	{
		mixer_set_curr(i);
		numchan = mixer_get_num_chan();
		NumChan[i] = numchan;
		for(j=0; j<numchan; j++)
		{
			chan_set_curr(j);
			n = strlen(chan_get_label());
			if( n > LabelWidth[i] )
			{
				n = CLAMP(n, n, LABELMAX+1);
				LabelWidth[i] = n;
			}
		}

		/* Allocate the balance arrays for every available mixer */
		/* okay, this just is sucky. I could store the balances 
		 * in one integer array, with -100 as left and 100 as right.
		 * But I'm lazy, and I will do this the way it used to be */
		S_Vol[i] = emalloc(sizeof(int) * numchan );
		LeftBal[i] = emalloc(sizeof(int) * numchan );
		RightBal[i] = emalloc(sizeof(int) * numchan );
	}
	mixer_set_curr(0);
	CurrMixer = 0;
	initvolumes(M_SAVE);
	loop();

	for(i=mixer_get_num_mix()-1; i>=0; i--)
	{
		free(S_Vol[i]);
		free(LeftBal[i]);
		free(RightBal[i]);
	}
	
	return 0;
}

/* the main loop */
static int loop(void)
{
	int ch, curr_chan; 
	int menudir, curr_menu=0;
	int slidedir;
	int change; 
	int left, right, max;
	int extradir;
	int quit=0;

	curr_chan = 0;
	chan_set_curr(curr_chan);
	while(1)
	{
		/* check if the window size has changed, this is called
		 * at least once in the beginning */
		if( (SCols != COLS) || (SLines != LINES) )
		{
			redrawscreen();
			drawpos(curr_menu, 1);
			SCols = COLS;
			SLines = LINES;
		}
		menudir = slidedir = extradir = 0;
		/* process input */
		if( (ch = getch()) != ERR )
		{
			switch(ch)
			{
			case 'j':
			case CTRL('n'):
			case KEY_DOWN:
				menudir = 1; break;
			case 'k': 
			case CTRL('p'):
			case KEY_UP:
				menudir = -1; break;
			case 'h':
			case CTRL('b'):
			case KEY_LEFT:
				slidedir = -1; break;
			case 'l':
			case CTRL('f'):
			case KEY_RIGHT:
				slidedir = 1; break;
			case ',':
			case '[':extradir = -(VOL_MAX/5); break;
			case '.':
			case ']':extradir = VOL_MAX/5; break;
			case ' ':togglerec(); break;
			case CTRL('l'): redrawscreen(); break;
			case 'H':help(); break; 
			case 'L':settings(SET_LOAD); break;
			case 'S':settings(SET_SAVE); break;
			case 'e':manualset(); break;
			case 'f':settings(SET_CONFIG); break;
			case 'x':mute(M_EXCLUDE, curr_chan); 
				 break;
			case 'm':mute(M_MUTE, -1); break;
			case 'n':cyclemixer(); break;
			case 'q':quit = 1; break;
			default: break;
			}
			if( quit == 1 )
				break;
			
			/* yes hack */
			ownupdate = 1;
		}
		/* check if mixer has been updated */
		if( mixer_get_status() == 1 )
		{
			/* if we didn't cause the update, 
			 * no need to draw again */
			if( ownupdate == 0 )
			{
				initvolumes(M_NOSAVE);
				drawall();
			}
		}
		ownupdate = 0;

		/* erase the old menu position */ 
		drawpos(curr_menu, 0);
		
		curr_chan += menudir;
		curr_chan = CLAMP(curr_chan, 0, NumChan[CurrMixer]-1);
		/* this is just stupid .. TODO: implement scrolling of
		 * channels if too small to fit on screen */
		if( curr_chan > (LINES - 6) )
			curr_chan = LINES - 6;

		chan_set_curr(curr_chan);

		/* change the menu .. needs to be here for
		 * drawing the menupos */
		if( ch == '\t' )
			curr_menu = ~curr_menu;

		/* draw the menu position */
		drawpos(curr_menu, 1);
#ifdef UMIX_DEBUG
		mvprintw(20, 2, "loop: curr_chan=%2d curr_mixer=%2d", 
			curr_chan, CurrMixer);
#endif
		if( slidedir != 0 || extradir != 0 )
		{
			/* the volume menu */
			if( curr_menu == 0 )
			{
				change = slidedir * VOLCHG;
				change += extradir;
				chan_inc_vol(change);
				drawslider();
			}	
			/* balance menu */
			else if( chan_is_stereo() )
			{
				left = LeftBal[CurrMixer][curr_chan];
				right = RightBal[CurrMixer][curr_chan];
				max = MAX(left, right);

				if( extradir == 0 )
					change = (max/BALWIDTH) + 1;
				else
				{
					slidedir = (extradir > 1) ? 1:0;
					change = max;
				}
				/* left and right are in the range 0 .. max */
				if( left > right )
				{
					right = (slidedir == 1) ?
						right + change : right - change;
				}
				else if( right > left )
				{
					left = (slidedir == 1) ?
						left - change : left + change;
				}
				/* if the current pos is center */
				else if( left == right )
				{
					if( slidedir == 1 )
						left -= change;
					else
						right -= change;
				}
				left = CLAMP(left, VOL_MIN, VOL_MAX);
				right = CLAMP(right, VOL_MIN, VOL_MAX);

				LeftBal[CurrMixer][curr_chan] = left;
				RightBal[CurrMixer][curr_chan] = right;

				drawbalance();
			}
			writevol();
#ifdef UMIX_DEBUG
			mvprintw(22, 2, "loop: balance is %3d:%3d",
				LeftBal[CurrMixer][curr_chan],
				RightBal[CurrMixer][curr_chan]);
			mvprintw(23, 2, "loop: volume is %3d:%3d",
				chan_get_lr(CHAN_LEFT), 
				chan_get_lr(CHAN_RIGHT));
#endif
		}
		refresh();
		napms(20);
	}
	
	return 0;
}

/* calculates the actual left and right values from current volume and
 * stored balances, then sets the volume */
static void writevol(void)
{
	int right, left;
	int volume, ch;
	int bl, br; 
		
	ch = chan_get_curr();
	volume = chan_get_vol();
	bl = LeftBal[CurrMixer][ch];
	br = RightBal[CurrMixer][ch];

	/* calculate the real level from the volume and balance */
	left = ((volume * bl)/VOL_MAX);	
	right = ((volume * br)/VOL_MAX);

	ownupdate = 1;
	
	chan_set_lr(left, right);
}

/* toggles the recording source between recsrc and play */
static void togglerec(void)
{
	if( chan_is_record() )
	{
		if( chan_is_recsrc() )
			chan_set_rec(CHAN_PLAY);
		else
			chan_set_rec(CHAN_RECSRC);
	}
}

/* update status for all channels */
static void updatestat(void)
{
	int i, orig_chan;

	orig_chan = chan_get_curr();
	for(i=0; i<NumChan[CurrMixer]; i++)
	{
		chan_set_curr(i);
		drawstatus();
	}
	chan_set_curr(orig_chan);
}

/* cycles through available mixer devices */
static void cyclemixer(void)
{
	int old = CurrMixer;
	int num_mix = mixer_get_num_mix();

	if( CurrMixer+1 < num_mix )
		CurrMixer++;
	else if( CurrMixer+1 >= num_mix )
		CurrMixer = 0;

	if( CurrMixer != old )
	{
		mixer_set_curr(CurrMixer);
		initvolumes(M_NOSAVE);
		redrawscreen();
	}
}

/* inits the vol and bal tables from the values in the mixer */
static void initvolumes(int save)
{
	int i;
	int left, right;
	int orig_chan;

	orig_chan = chan_get_curr();
	for(i=0; i<NumChan[CurrMixer]; i++)
	{
		chan_set_curr(i);
		left = chan_get_lr(CHAN_LEFT);
		right = chan_get_lr(CHAN_RIGHT);
		/* save previous maximum to S_VOL */
		if( save == M_SAVE )
			S_Vol[CurrMixer][i] = chan_get_vol();

		/* balance levels are in the range 0 - VOL_MAX */
		if( left != right )
		{
			if( left > right )
			{
				right <<= 12;
				LeftBal[CurrMixer][i] = VOL_MAX;
				RightBal[CurrMixer][i] = (VOL_MAX * 
					(right/left)) >> 12;
			}
			else
			{
				left <<= 12;
				LeftBal[CurrMixer][i] = (VOL_MAX * 
					(left/right)) >> 12;
				RightBal[CurrMixer][i] = VOL_MAX;
			}
		}
		else
		{
			LeftBal[CurrMixer][i] = VOL_MAX;
			RightBal[CurrMixer][i] = VOL_MAX;
		}
	}
	chan_set_curr(orig_chan);
}

/* Mutes and demutes either all channels or one channel */
static void mute(int action, int ch)
{
	int i;
	int orig_chan;
	char *msg = NULL;

	/* all channels */
	if( ch == -1 )
	{
		orig_chan = chan_get_curr();
		for(i=0; i<NumChan[CurrMixer]; i++)
		{
			chan_set_curr(i);
			/* if channel not already mute, mute */
			if( chan_is_mute() == 0 )
			{
				S_Vol[CurrMixer][i] = chan_get_vol();
				chan_set_vol(0);
			}
			/* else restore old level */
			else
				chan_set_vol(S_Vol[CurrMixer][i]);
			
			writevol();
			drawslider();
		}
		chan_set_curr(orig_chan);
		msg = _("Mute");
	}
	else
	{
		/* exclude channel */
		if( action == M_EXCLUDE )
		{
			/* mute */
			if( chan_is_mute() == 0 )
			{
				S_Vol[CurrMixer][ch] = chan_get_vol();
				chan_set_vol(0);
				msg = _("Exclude");
			}
			/* restore */
			else
			{
				chan_set_vol(S_Vol[CurrMixer][ch]);
				msg = _("Include");
			}
		}
		writevol();
		drawslider();
	}
	drawborder(BORD_LOWER, msg);
}

/* draws the menu position markers */
static void drawpos(int menu, int status)
{
	char start, end;
	int x1, x2;
	char buf[LABELMAX];
	int ch = chan_get_curr();

	if( status == 1 )
	{
		start = '>';
		end = '<';
	}
	else
		start = end = ' ';
	/* what menu are we in ? */
	if( menu == 0 )
	{
		x1 = VOLPOSX1;
		x2 = VOLPOSX2;		
	}
	else
	{
		x1 = BALPOSX1;
		x2 = BALPOSX2; 
	}
	
	attron(A_BOLD);
	mvaddch(BEGY+ch, x1, start);
	mvaddch(BEGY+ch, x2, end);
	attroff(A_BOLD);
	
	if( status == 1 )
		attron(A_BOLD);
	snprintf(buf, LabelWidth[CurrMixer]+1, "%s", chan_get_label());
	mvaddstr(BEGY+ch, BEGX, buf);
	if( status == 1 ) 
		attroff(A_BOLD);
}

/* draws channel status information */
static void drawstatus(void)
{
	char rec;
	int ch = chan_get_curr();

	if( chan_is_record() )
	{
		if( chan_is_recsrc() )
			rec = 'R';
		else
			rec = 'P';
		
		attron(A_BOLD);
		mvprintw(BEGY+ch, STATBEGX, "%c", rec);
		attroff(A_BOLD);
	}
}

/* draws the volume level and slider */
static void drawslider(void)
{
	int i;
	int sliderwidth, volume;
	int y;
	int x = VOLBEGX;

	y = BEGY + chan_get_curr();

	volume = chan_get_lr(CHAN_MAX);
	sliderwidth = (volume * VOLWIDTH) / VOL_MAX;
	mvprintw(y, x, "%3d", volume);
	x += NUMWIDTH + 2;
	
	/* draw the characters indicating empty */
	for(i=0; i<VOLWIDTH; i++)
		mvaddch(y, x+i, CHAR_EMPTY);
	mvaddch(y, x, CHAR_DIV);
	mvaddch(y, x+(VOLWIDTH-1), CHAR_DIV);

	/* draw the slider */
	move(y, x);
	attron(A_BOLD);
	hline(CHAR_SLIDER, sliderwidth);
	attroff(A_BOLD);
}

/* draws the balance slider */
static void drawbalance(void)
{
	int left, right;
	int pos;
	char mark;
	int ch;
	int y;
	int x = BALBEGX;

	ch = chan_get_curr();
	y = BEGY + ch;
	
	left = LeftBal[CurrMixer][ch];
	right = RightBal[CurrMixer][ch];
	
	if( left != right )
	{
		static int max;
		
		max = MAX(left, right);
		/* count the position of the balance marker */
		if( left > right )
			pos = ((BALWIDTH/2) * right) / max;
		else
			pos = BALWIDTH - ((BALWIDTH/2) * left) / max;
	}
	else
		pos = BALWIDTH/2;
#ifdef UMIX_DEBUG
	mvprintw(26, 2, "drawbalance: pos=%2d %3d:%3d", 
		pos, left, right);
#endif
	/* draw the base */
	move(y, x);
	hline(CHAR_EMPTY, BALWIDTH);
	/* draw the start, end and center markers */
	mvaddch(y, x, CHAR_DIV);
	mvaddch(y, x+BALWIDTH, CHAR_DIV);
	mvaddch(y, x+BALWIDTH/2, CHAR_DIV);
	
	if( chan_is_stereo() )
		mark = CHAR_SLIDER;
	else
		mark = CHAR_MONO;

	/* draw the marker */
	attron(A_BOLD);
	mvaddch(y, x+pos, mark);
	attroff(A_BOLD);
}

/* draws the label, volume, balance slider and status for every channel */ 
static void drawall(void)
{
	int y = BEGY;
	int i;
	int origch;
	char buf[LABELMAX];

	origch = chan_get_curr();
	for(i=0; i<NumChan[CurrMixer]; i++)
	{
		chan_set_curr(i);
		snprintf(buf, LabelWidth[CurrMixer]+1, "%s", 
			chan_get_label());
		mvaddstr(y, BEGX, buf);
		drawslider();
		drawbalance();
		drawstatus();
		y++;
		if( y > (LINES - 4) )
			break;
	}
	chan_set_curr(origch);
}

/* draws the upper or lower border with optional string centered */
static void drawborder(int where, const char *str)
{
	int y, x;

	y = (where == BORD_UPPER ) ? 0 : LINES-2;

	attron(A_REVERSE);
	move(y, 0);
	hline(' ', COLS);
	if( str != NULL )
	{
		x = COLS/2 - strlen(str)/2;
		mvaddstr(y, x, str);
	}
	attroff(A_REVERSE);
}

/* refreshes the screen */
static void redrawscreen(void)
{
	char buf[LINEWIDTH]; 

	clear();
	snprintf(buf, sizeof(buf), "%s v%s (%s::%s) %d/%d", NAME_STR, VERSION, 
		mixer_get_name(), mixer_get_path(), mixer_get_curr()+1,
		mixer_get_num_mix());
	drawborder(BORD_UPPER, buf);
	snprintf(buf, sizeof(buf), _("Press H for help"));
	drawborder(BORD_LOWER, buf);
	drawall();
	refresh();
}

/* loads or saves settings depending on action */
static void settings(int action)
{
	int retval;
	char buf[BUFLEN];
	
	if( action == SET_LOAD )
	{
		retval = loadsettings(globalopts.cfgpath, SET_MIXER);
		if( retval == ESETERR )
		{
			snprintf(buf, sizeof(buf), _("%s: %s"),
				globalopts.cfgpath, strerror(errno));
		}
		else if( retval == ESETINVALID )
		{
			snprintf(buf, sizeof(buf), 
			_("Invalid config file"));
		}
		else if( retval == ESETNOTFOUND )
		{
			snprintf(buf, sizeof(buf),
			_("Volumes for current mixer not found in config"));
		}
		else if( retval == ESETBINARY )
		{
			snprintf(buf, sizeof(buf),
			_("Binary config file detected, not loading"));
		}
		else
		{
			snprintf(buf, sizeof(buf), 
				_("Settings loaded from %s"), 
				globalopts.cfgpath);
			initvolumes(M_SAVE);
			drawall();
		}
	}
	else if( action == SET_SAVE )
	{
		if( (savesettings(globalopts.cfgpath) == -1 ) )
		{
			snprintf(buf, sizeof(buf), _("Failed saving %s: %s"), 
				globalopts.cfgpath, strerror(errno));
		}
		else
			snprintf(buf, sizeof(buf), _("Settings saved to %s"), 
				globalopts.cfgpath);
	}
	else
	{
		/* falls to inputting the path for the config */
		static int n;

		input(buf, _("Enter config path: "), 64);
		n = strlen(buf);
		if( n > 0 && n < COLS )
		{
			strncpy(globalopts.cfgpath, buf, 
				sizeof(globalopts.cfgpath));
			snprintf(buf, sizeof(buf), 
				_("Config path set to \"%s\""), 
				globalopts.cfgpath);
		}
		else
			snprintf(buf, sizeof(buf), 
				_("Config path \"%s\" not changed"), 
				globalopts.cfgpath);
			
		drawborder(BORD_LOWER, buf);
		return;
	}
	/* draw the message */
	drawborder(BORD_LOWER, buf);
}

/* gets input from the keyboard to inbuf, with optional prompt */
static void input(char *inbuf, const char *prompt, int limit)
{
	if( limit <= 1 )
		return;
	nodelay(stdscr, FALSE);
	refresh();
	if( prompt != NULL )
		mvaddstr(LINES-1, 0, prompt); 	
	echo();
	getnstr(inbuf, limit-1);
	noecho();
	hline(' ', COLS);
	nodelay(stdscr, TRUE);
}

static void manualset()
{
	char inbuf[TINYSTRLEN/2];
	char buf[STRLEN];
	int ch = chan_get_curr();

	snprintf(buf, sizeof(buf), _("Enter volume for channel %d: "), ch);
	input(inbuf, buf, sizeof(inbuf));
	if( parsechanopt(inbuf) == -1 )
	{
		snprintf(buf, sizeof(buf), 
			_("Invalid value for channel %d: %s"), ch, inbuf);
		drawborder(BORD_LOWER, buf);
	}
	else
		drawborder(BORD_LOWER, NULL);

	initvolumes(M_NOSAVE);
	drawslider();
	drawbalance();
	updatestat();
}

static void help(void)
{
	clear();
	drawborder(BORD_UPPER, _("Umix help"));
	drawborder(BORD_LOWER, _("Press any key to return"));
	mvaddstr(BEGY, BEGX,
		_(
		"j / ^n / arrow down\tMove down\n"
		"k / ^p / arrow up\tMove up\n"
		"h / ^b / arrow left\tMove left\n"
		"l / ^f / arrow right\tMove right\n"
		"] / ,\t\t\tIncrease volume by fifth of maximum amount\n"
		"[ / .\t\t\tDecrease volume by fifth of maximum amount\n"
		"Tab\t\t\tToggle between volume and balance menus\n"
		"Space\t\t\tToggle play / record source\n"
		"L\t\t\tLoad settings\n"
		"S\t\t\tSave settings\n"
		"f\t\t\tSpecify the config path for settings\n"
		"e\t\t\tEnter volume manually for current channel\n"
		"m\t\t\tMute all channels, undo if already mute\n"
		"n\t\t\tCycle through available mixers\n"
		"x\t\t\tSame as m but with only with the current channel\n"
		"H\t\t\tThis help\n"
		"q\t\t\tExit the program\n"
		)
	);
	refresh();
	nodelay(stdscr, FALSE);
	getch();
	nodelay(stdscr, TRUE);
	redrawscreen();
}

static void initscreen(void)
{
	initscr();
	keypad(stdscr, TRUE);
	nodelay(stdscr, TRUE);
	nonl();
	cbreak();
	noecho();
}

/* deinit ncurses */
void ui_ncurs_exit(void)
{
	clear();
	refresh();
	endwin();
#ifdef UMIX_DEBUG
	errormsg("ui_ncurs_exit: Ncurses UI successfully deinitialized");
#endif 
}
