/*
 * umix.c: 
 * The main function and command-line UI for Umix.
 * Main parses command-line options, initializes the mixer 
 * and then calls the UI. The command-line UI does not
 * support multiple devices in one session.
 *
 * Copyright (C) 2001 Sakari Lehtonen <sakari@ionstream.fi>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include "umix_common.h"
#ifdef HAVE_GETOPT_LONG
  #include <getopt.h>
#else
  #include <unistd.h>
#endif
#include "plugin.h"
#include "ui.h"

/* global options */
umix_cfg globalopts =
{
	DEF_DEVICE_PATH,
	DEF_CONFIG_PATH,
	DEF_UI,
	DEF_DRIVER,
	DEF_PLUGIN_DIR
};

/* function prototypes */
static void query(void);
static int printlevels(void);
static void usage(int status);
static void version(void);
static void showglobals(void);
static int setall(const char *optarg);

/* the name umix is called with */
const char *program_name;

int main(int argc, char *argv[])
{
	int optch=0; 
	int interactive=0;
	int opt_q=0, opt_l=0, opt_s=0, opt_g=0, opt_i=0;
        int i, chnum=0;
	char *home;

        const char *optstr = "d:r:f:p:u:hlLqvsSig";
#ifdef HAVE_GETOPT_LONG
        const struct option longopts[] =
        {
		{ "device",	1, 0, 'd' },
		{ "driver",	1, 0, 'r' },	
                { "file",       1, 0, 'f' },
		{ "plugindir",	1, 0, 'p' },
		{ "globals",	0, 0, 'g' },
                { "help",       0, 0, 'h' },
                { "load",       0, 0, 'l' },
                { "query",      0, 0, 'q' },
                { "save",       0, 0, 's' },
		{ "ui",		1, 0, 'u' },
                { "version",    0, 0, 'v' },
		{ "interactive",0, 0, 'i' },
                { NULL,         0, 0, 0 }
        };
#endif /* HAVE_GETOPT_LONG */

#ifdef ENABLE_NLS
	/* internationalization */
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);
#endif
	program_name = argv[0];

	/* check if $HOME/.umixrc exists, and use as the default
	 * config if found */
	if( (home = getenv("HOME")) != NULL )
	{
		static char buf[STRLEN];
		static struct stat sbuf;

		snprintf(buf, sizeof(buf), "%s/.umixrc", home);
		if( stat(buf, &sbuf) == 0 )
		{
			strncpy(globalopts.cfgpath, buf,
				sizeof(globalopts.cfgpath));
		}
	}
	
	/* try to load global settings if not loading or saving to stdin */
	if( globalopts.cfgpath[0] != '-' && globalopts.cfgpath[1] != '\0' )
		loadsettings(globalopts.cfgpath, SET_GLOBAL);
	
	/* parse command-line options */
#ifdef HAVE_GETOPT_LONG
	while( (optch = getopt_long(argc, argv, optstr, longopts, NULL)) != -1 )
	{
#else
	while( (optch = getopt(argc, argv, optstr)) != -1 )
        {
#endif
		switch(optch)
		{
		case 'd': strncpy(globalopts.devpath, optarg,
				  sizeof(globalopts.devpath));
			  interactive = 1;
			  break;
		case 'f': strncpy(globalopts.cfgpath, optarg,
				  sizeof(globalopts.cfgpath));
			  interactive = 1;
			  break;
		case 'g': opt_g = 1; break;
		case 'h': usage(EXIT_SUCCESS); break;
		case 'u': strncpy(globalopts.ui, optarg,
				  sizeof(globalopts.ui));
			  /* fallthrough */
		case 'i': opt_i = 1; break;
		case 'L': /* compatibility */
		case 'l': opt_l = 1; break;
		case 'p': strncpy(globalopts.plugindir, optarg,
				  sizeof(globalopts.plugindir));
			  interactive = 1;
			  break;
		case 'q': opt_q = 1; break;
		case 'r': strncpy(globalopts.driver, optarg,
				  sizeof(globalopts.driver));
			  interactive = 1;
			  break;
		case 'S': /* compatibility */
		case 's': opt_s = 1; break;
		case 'v': version(); exit(EXIT_SUCCESS); break;
		default: usage(EXIT_FAILURE); break;
		}
        };
	
	/* this is just a way to check if umix was called with only one
	 * argument eg `umix foo` .. maybe a better way ? */
	if( argc == 2 && optind == 1 )
	{
		errormsg(_("nothing to do!"));
		usage(EXIT_FAILURE);
	}

	if( init_plugins() == -1 )
	{
		errormsg(_("failed initializing plugins"));
		return -1;
	}
	if( init_mixers() == -1 )
	{
		errormsg(_("failed initializing mixer devices"));
		return -1;
	}

	/* process all command-line options conserning channels 
	 * channel format is the following :
	 *
	 * name volume|left:right[-|+]p|r
	 *
	 * where name is the name of the channel. volume is either 
	 * volume for both channels, or individual left and right values 
	 * separated with a ':'. the last char sets the recording 
	 * state of the channel to play 'p' or record source 'r'. */
	for(i=optind; (i+1) < argc; i++)
	{
		if( (strcasecmp(argv[i], "all")) == 0 )
		{
			i++;
			setall(argv[i]);
		}
		/* get the right channel number for the string argv[i] eg.
		 * "pcm" would return 1 with my system */
		else if( (chnum = mixer_opt_to_chan_num(argv[i])) != -1 )
		{
			i++;
			chan_set_curr(chnum);
			if( parsechanopt(argv[i]) == -1 )
				errormsg(_("invalid value for channel %s: %s"),
					chan_get_name(), argv[i]);
		}
		else
			errormsg(_("invalid channel name: %s"), argv[i]);
	}
	/* if any channel options are given, do not start interactive ui */
	if( optind != argc )
		interactive = 0;

	/* query settings */
	if( opt_q == 1 )
	{
		query();
		/* no point in both querying levels and starting 
		 * the interactive ui */
		interactive = 0;
	}
	/* show global settings */
	if( opt_g == 1 )
	{
		showglobals();
		interactive = 0;
	}
	/* load settings */
	if( opt_l == 1 )
	{
		static int retval;
		printf(_("Loading mixer settings from %s\n"),
			globalopts.cfgpath);
		retval = loadsettings(globalopts.cfgpath, SET_MIXER);

		if( retval == ESETERR )
			errorstr(_("failed loading settings from %s"), 
				globalopts.cfgpath);
		else if( retval == ESETINVALID )
			errormsg(_("%s: invalid config file"),
				globalopts.cfgpath);
		else if( retval == ESETNOTFOUND )
			errormsg(_("%s: mixer settings for %s not found")
				,globalopts.cfgpath, globalopts.devpath);
		else if( retval == ESETBINARY )
			errormsg(_("%s: binary file detected, not loading"), 
				globalopts.cfgpath);
	
		interactive = 0;
	}
	/* save settings */
	if( opt_s == 1 )
	{
		printf(_("Saving settings to %s\n"), globalopts.cfgpath);
		if( savesettings(globalopts.cfgpath) == -1 )
		{
			errorstr(_("failed saving settings to %s"), 
				globalopts.cfgpath);
		}
		
		interactive = 0;
	}

	/* opt_i forces to interactive */
	if( opt_i == 1 )
		interactive = 1; 

	if( argc < 2 || interactive == 1 ) 
	{
		if( strcasecmp(globalopts.ui, "cli") == 0 )
			query();
		else
		{
			/* no sense in saving or loading settings 
			 * from stdin/stdout in interactive UI's, replace
			 * config path with the default one */
			if( globalopts.cfgpath[0]=='-' && 
				globalopts.cfgpath[1]=='\0')
			{
				strncpy(globalopts.cfgpath, DEF_CONFIG_PATH,
					sizeof(globalopts.cfgpath));
			}
			if( ui_init(globalopts.ui) == -1 )
			{
				errormsg(_("try setting ui=cli in config file"
				      ", or use `%s -q'"), program_name);
			}
		}
	}

	close_mixers();
	/*
	test(0);
	*/
	return 0;
}

static void query(void)
{
	int i;
	for(i=0; i<mixer_get_num_mix(); i++)
	{
		mixer_set_curr(i);
		printlevels();
	}
}

/* parse optarg for all channels */
static int setall(const char *optarg)
{
        int i;

        for(i=mixer_get_num_chan()-1; i>=0; i--)
        {
		chan_set_curr(i);
                if( parsechanopt(optarg) == -1 )
			return -1;
        }

        return 0;
}

/* prints the levels and status for all channels on current mixer */
static int printlevels(void)
{
        int i;

	printf("%s::%s::%s\n", mixer_get_name(), mixer_get_path(),
		mixer_get_driver());
        for(i=0; i<mixer_get_num_chan(); i++)
        {
		chan_set_curr(i);
		printf("%-10s %3d:%3d ", chan_get_name(), 
			chan_get_lr(CHAN_LEFT), chan_get_lr(CHAN_RIGHT));

		if( chan_is_stereo() )
			printf(_("Stereo"));
		else
			printf(_("Mono  "));
		putchar(' ');

		if( chan_is_recsrc() )
			printf(_("Record Source"));
		else if( chan_is_record() )
			printf(_("Recordable"));
		putchar('\n');
        }

        return 0;
}

static void showglobals(void)
{
	printf(_("Current global settings:\n"
		"Device path\t   %s\n"
	       	"Mixer driver\t   %s\n"
	       	"Config path\t   %s\n"
	       	"Plugin directory   %s\n"
	       	"User interface\t   %s\n"),
	       	globalopts.devpath, globalopts.driver, globalopts.cfgpath,
	       	globalopts.plugindir, globalopts.ui);
}

static void usage(int status)
{
	if( status != 0 )
	{
		fprintf(stderr, _("try `%s -h' for more information\n"),
			program_name);
	}
	else
	{
        printf(_("Usage: %s [channame volume|left:right][R|P][+|-] [options]\n"
	     "\nYou can get all available channel names with the '-q' option"
	     "\n\nAdditional options:\n"
             "-d, --dev=PATH\t\tSpecify device used for mixing\n"
	     "\t\t\t(defaults to %s)\n"
	     "-r, --driver=DRIVER\tSpecify the mixer driver used\n"
	     "\t\t\t(defaults to %s)\n"
             "-f, --file=PATH\t\tSpecify file used to read and save settings\n"
	     "\t\t\t(defaults to %s)\n"
	     "-p, --plugindir=PATH\tSpecify the path to load plugins from\n"
	     "\t\t\t(defaults to %s)\n"
             "-h, --help\t\tDisplay this help and exit\n"
	     "-i, --interactive"
	     "\tStart interactive user interface after processing\n"
	     "\t\t\tcommand-line options (if available)\n"
             "-l, --load\t\tLoad mixer settings from a file\n"
             "-q, --query\t\tPrint channel settings and available\n"
	     "\t\t\tchannel names\n"
	     "-g, --globals\t\tPrint current global options\n"
	     "-s, --save\t\tSave mixer settings to a file\n"
	     "-u, --ui\t\tSelect the UI to be used\n"
             "-v, --version\t\tDisplay version information and exit\n"),
		program_name, DEF_DEVICE_PATH, DEF_DRIVER, DEF_CONFIG_PATH, 
		DEF_PLUGIN_DIR);
	}

	exit(status);
}

static void version(void)
{
        printf(_("%s version %s\n"
	"Copyright (C) 2001 Sakari Lehtonen <sakari@ionstream.fi>\n"
	"\nThis is free software; see the source for copying conditions. " 
	"There is NO\nwarranty; not even for MERCHANTABILITY or FITNESS FOR A" 
	" PARTICULAR PURPOSE.\n"), NAME_STR, VERSION);
}

#ifdef UMIX_DEBUG
/* tests most of the functions available, and shows how to do things */
int test(int mode)
{
	int i, j;
	int orig_vol;

	/* assumes that OSS is available */
	mixer_load_drv("oss");
	mixer_open("/dev/mixer");
	mixer_init();
	
	/* update check test */
	if( mode == 1 )
	{
		while( 1 )
		{
			usleep(250000);
			if( mixer_get_status() == 1 )
			{
				errormsg("mixer has been updated!");
				printlevels();
			}
		}

		mixer_close();
		return 0;
	}

#if 0
	/* hmm testing reverse order, is faster because calls
	 * mixer_get_num_mix() and mixer_get_num_chan() only once */
	for(j=mixer_get_num_mix()-1; j>=0; j--)
	{
		mixer_set_curr(j);
		errormsg("mixer=%d", mixer_get_curr());
		for(i=mixer_get_num_chan()-1; i>=0; i--)
		{
			chan_set_curr(i);
			errormsg("  chan=%d", chan_get_curr());
		}
	}
#endif

	printf("Total number of mixers = %d\n", mixer_get_num_mix());
	for(j=0; j<mixer_get_num_mix(); j++)
	{
		printf("Mixer #%d information\n", mixer_get_curr());
		printf("numchan = %d\n", mixer_get_num_chan());
		printf("name    = %s\n", mixer_get_name());
		printf("path    = %s\n", mixer_get_path());
		printf("driver  = %s\n", mixer_get_driver());

		printf("num name       label      lt  rt   s r a m\n");
		for(i=0; i<mixer_get_num_chan(); i++)
		{
			chan_set_curr(i);
			printf("%3d %-10s %-10s ", chan_get_curr(), 
				chan_get_name(), chan_get_label());
			printf("%3d %3d  %d %d %d %d\n", 
				chan_get_lr(CHAN_LEFT), chan_get_lr(CHAN_RIGHT),
				chan_is_stereo(), chan_is_recsrc(), 
				chan_is_record(), chan_is_mute());
		}

		chan_set_curr(0);
		chan_set_elem(CHAN_MAX);
		orig_vol = chan_get_vol();

		printf("Setting the first channel to 100\n");
		chan_set_vol(100);
		usleep(200000);

		printf("Setting balance to right\n");
		chan_set_lr(0, 100);
		usleep(200000);

		printf("Setting balance to left\n");
		chan_set_lr(100, 0);
		usleep(200000);

		printf("Setting balance back to center\n");
		chan_set_lr(100, 100);

		printf("Decreasing volume with 2\n");
		chan_inc_vol(-2);
		errormsg("left=%d right=%d", chan_get_lr(CHAN_LEFT),
			chan_get_lr(CHAN_RIGHT));
		usleep(200000);

		printf("Increasing volume with 2\n");
		chan_inc_vol(2);
		errormsg("left=%d right=%d", chan_get_lr(CHAN_LEFT),
			chan_get_lr(CHAN_RIGHT));
		usleep(200000);

		printf("Restoring old values: vol=%3d\n", 
			orig_vol);
		chan_set_vol(orig_vol);

		for(i=mixer_get_num_chan()-1; i>=0; i--)
		{
			chan_set_curr(i);
			if( chan_is_record() )
			{
				printf("Setting channel #%d as recsrc\n", i);
				chan_set_rec(CHAN_RECSRC);
				if( chan_is_recsrc() )
				{
					printf("Channel #%d (%s) is recsrc\n",
						i, chan_get_name());
				}
				else
					printf("Setting recsrc failed!\n");

				break;
			}
		}
	}

	query();
	mixer_close();

	return 0;
}
#endif
