/*
 * umix_common.c: 
 * Common functions for Umix. All UIs have to be able to call
 * these, and errors are handled by the caller.
 * 
 * Copyright (C) 2001 Sakari Lehtonen <sakari@ionstream.fi>
 *
 */

#include "umix_common.h"

static void parseglobal(char *buf);
static void parsemixer(char *buf);

/* Parses one line of channel options.
 * TODO: more refinement, some bogus characters will cause the volume to be 
 * parsed as 10000 or something like that, and thus setting the volume to the 
 * max value */
int parsechanopt(const char *optarg)
{
	int change=0;
	int left, right;
	int recstate;
	int origvol;
	int slidefound=0;
	size_t n;

	/* move local optarg pointer forwards so no whitespace will be
	 * parsed */
	while(*optarg == ' ' || *optarg == '\t')
		optarg++;
	/* check for increment or decrement */
	n = strlen(optarg)-1;
	if( optarg[n] == '+' || optarg[n] == '-' )
		change = (optarg[n] == '+') ? 1 : -1;

#ifdef UMIX_DEBUG
	errormsg("parsechanopt: optarg=%s change=%d", optarg, change);
#endif
	/* check if we are not just setting recording states, then optarg
	 * would only contain 'r' or 'p', and no volumes would be needed
	 * to set */
	if( isdigit(optarg[0]) )
	{
		if( strchr(optarg, ',') != NULL )
			slidefound = 1;

		/* if no ':' found, use same value for both */
		if( strchr(optarg, ':') == NULL )
		{
			left = atoi(optarg);
			right = left;
		}
		else
		{
			sscanf(optarg, "%3d:%3d", &left, &right);
			if( chan_is_stereo() == 0 )
				right = left;
		}
	}
	else
	{
		/* check for garbage */
		recstate = tolower(optarg[0]);
		if( recstate != 'r' && recstate != 'p' )
			return -1;
	}
	/* check for setting of record source or play */
	n = strcspn(optarg, "rRpP");
	recstate = tolower(optarg[n]);
	if( (recstate == 'r' || recstate == 'p') )
	{
		if( chan_is_record() )
		{
			if( recstate == 'r' )
				recstate = CHAN_RECSRC;
			else
				recstate = CHAN_PLAY;
			chan_set_rec(recstate);
		}
		/* P or R is the only argument for this channel, return */
		if( isdigit(optarg[0]) == 0 )
			return 0;
	}
	/* add or decrement from the old values ? */
	if( change != 0 )
	{
		chan_set_elem(ELEM_LEFT);
		origvol = chan_get_vol();
		left  = (change == 1 ) ? 
			origvol + left : 
			origvol - left;

		chan_set_elem(ELEM_RIGHT);
		origvol = chan_get_vol();
		right = (change == 1 ) ? 
			origvol + right : 
			origvol - right;
	}
	chan_set_lr(left, right);
#ifdef UMIX_DEBUG
	errormsg("parsechanopt: left=%d right=%d parsedleft=%d parsedright=%d",
		chan_get_lr(CHAN_LEFT), chan_get_lr(CHAN_RIGHT), left, right);
#endif
	return 0;
}

/* the mixer that is used when parsing mixer settings */
static int set_mixer = -1;

/* load settings from file defined in path */
int loadsettings(const char *path, int mode)
{
	FILE *fp;
	void (*parseline)(char *buf);
	char buf[LINEWIDTH];
	size_t n;
	int globalfound = 0, mixersfound = 0;

	/* lines are parsed depending on the mode */
	if( mode == SET_GLOBAL )
		parseline = parseglobal;
	else
		parseline = parsemixer;

	if( path[0] == '-' && path[1] == '\0' )
		fp = stdin;
	else if( (fp = fopen(path, "r")) == NULL )
		return ESETERR;

	while( fgets(buf, sizeof(buf), fp) != NULL )
	{
		/* try to detect binary files .. this _should_ detect 
		 * garbage/binary .. TODO: check */
		if( isascii(buf[0]) == 0 )
			return ESETBINARY;
		/* check for comments */
		n = strcspn(buf, "#");
		if( n < 2 )
			continue;

		/* remove newline */
		n = strlen(buf) - 1;
		if( buf[n] == '\n' || buf[n] == '\r' )
			buf[n] = '\0';

		if( strncmp(buf, "[global]", 8) == 0 )
		{
			globalfound = 1;
			if( (mode == SET_MIXER) && (mixersfound == 1) )
				break;
			continue;
		}
		else if( strncmp(buf, "[mixers]", 8) == 0 )
		{
			mixersfound = 1;
			if( (mode == SET_GLOBAL) && (globalfound == 1) )
				break;
			continue;
		}
		/* parse lines if the mode is right and beginning found */
		if( ((mode == SET_GLOBAL) && globalfound == 1) ||
		    ((mode == SET_MIXER) && mixersfound == 1 ))
			parseline(buf);
	}

	if( globalfound == 0 && mixersfound == 0 )
		return ESETINVALID;

	if( (mode == SET_MIXER) && (set_mixer == -1) )
		return ESETNOTFOUND;

	if( fp == stdin )
		return 0;
	else fclose(fp);

	return 0;
}

/* parses one line in the [global] section of config file */
static void parseglobal(char *buf)
{
	size_t n;

	/* hmm yes there might be a better way to do this .. but this is
	 * sufficient and simple enough */
	n = strcspn(buf, "=");
	buf[n++] = '\0';
	if( strcmp(buf, "device") == 0 )
		strncpy(globalopts.devpath, buf + n,
			sizeof(globalopts.devpath));
	else if( strcmp(buf, "config") == 0 )
		strncpy(globalopts.cfgpath, buf + n,
			sizeof(globalopts.cfgpath));
	else if( strcmp(buf, "ui") == 0 )
		strncpy(globalopts.ui, buf + n,
			sizeof(globalopts.ui));
	else if( strcmp(buf, "driver") == 0 )
		strncpy(globalopts.driver, buf + n,
			sizeof(globalopts.driver));
	else if( strcmp(buf, "plugindir") == 0 )
		strncpy(globalopts.plugindir, buf + n,
			sizeof(globalopts.plugindir));
}

/* parses one line in the [mixers] section of config file */
static void parsemixer(char *buf)
{
	size_t n;
	int i;

	/* these aren't really required in the config file .. they are
	 * just helping readability */
	if( *buf == '{' || *buf == '}' )
		return;
	/* remove leading whitespace */
	while( *buf == ' ' || *buf == '\t' )
		buf++;

	/* find the mixer */
	if( strncmp(buf, "mixer=", 6) == 0 )
	{
		buf += 6;
		for(i=mixer_get_num_mix()-1; i>=0; i--)
		{
			/* compare device paths to find out the right 
			 * mixer */
			mixer_set_curr(i);
			if( strcmp(buf, mixer_get_path()) == 0 )
			{
				/* found */
				set_mixer = i;
				break;
			}
		}
	}
	else if( set_mixer != -1 )
	{
		static int chnum;

		mixer_set_curr(set_mixer);
		n = strcspn(buf, " ");
		buf[n] = '\0';
		if( (chnum = mixer_opt_to_chan_num(buf)) != -1 )
		{
			chan_set_curr(chnum);
			parsechanopt(buf+n+1);
		}
	}
}

/* save mixer settings to a file defined in path. */
int savesettings(const char *path)
{
	FILE *fp;
	int i, j, n;
	char record;

	if( path[0] == '-' && path[1] == '\0' )
		fp = stdout;
	else if( (fp = fopen(path, "w+")) == NULL )
		return ESETERR;

	fprintf(fp, "# Umix config file\n"
		    "[global]\n"
		    "device=%s\n"
		    "config=%s\n"
		    "plugindir=%s\n"
		    "ui=%s\n"
		    "driver=%s\n\n",
		    globalopts.devpath, globalopts.cfgpath,
		    globalopts.plugindir, globalopts.ui, globalopts.driver);
	fprintf(fp, "[mixers]\n");
	
	/* hmhm. we do this so we dont have to call mixer_get_num_mix()
	 * over every iteration, and this is faster on some processors too */
	for(i=mixer_get_num_mix()-1; i>=0; i--)
	{
		mixer_set_curr(i);
		/* print the name of the mixer and mixer driver name
		 * just to add readability */
		fprintf(fp, "# %s, %s\n", mixer_get_name(), mixer_get_driver());
		fprintf(fp, "mixer=%s\n{\n", mixer_get_path());

		/* we must do this to get the channels in right order */
		n = mixer_get_num_chan();
		for(j=0; j<n; j++)
		{
			chan_set_curr(j);
			if( chan_is_recsrc() )
				record = 'R';
			else if( chan_is_record() )
				record = 'P';
			else
				record = ' ';
			fprintf(fp, "\t%s %d:%d %c\n", 
				chan_get_name(), chan_get_lr(CHAN_LEFT),
				chan_get_lr(CHAN_RIGHT), record);
		}
		fprintf(fp, "}\n");
	}
	
	/* we dont want to close stdout .. */
	if( fp == stdout )
		return 0;
	else fclose(fp);

	return 0;
}

/* add nameval to growing array tab */
int addnv(nvtab *tab, nameval newname)
{
	nameval *nvp;

	/* first time called on tab */
	if( tab->nv == NULL )
	{	
		tab->nv = malloc(NVINIT * sizeof(nameval));
		if( tab->nv == NULL )
			return -1;
		tab->max = NVINIT;
		tab->nval = 0;
	}
	else if( tab->nval >= tab->max )
	{
		/* reallocate nv by multiplier of NVGROW */
		nvp = erealloc(tab->nv, (NVGROW*tab->max) * sizeof(nameval));
		if( nvp == NULL )
			return -1;
		tab->max *= NVGROW;
		tab->nv = nvp;
	}
	tab->nv[tab->nval] = newname;
	
	return tab->nval++;
}

/* search and return nameval by name from tab */
nameval *lookupbyname(nvtab *tab, const char *name)
{
	int i;

	for(i=0; i<tab->nval; i++)
	{
		if( strcmp(tab->nv[i].name, name) == 0 )
			return tab->nv;
	}

	return NULL;
}
